/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2009, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file dist_list.h
 *
 * @brief The header for distributed lists.
 */
#ifndef __DIST_LIST_H__
#define __DIST_LIST_H__
namespace sketo {

  /**
   * @brief A class for distributed list (one-dimensional array).
   *
   */
  template <typename A>
  class dist_list {
    friend class _impl::ls_adpt;
    
  public:
    typedef A element_type;

  private:
    _impl::dist_list_buffer<A> *data;

  public:
    dist_list();

    /** @brief A constructor that allocates distributed space */
    dist_list(int size);

    /** @brief A constructor that allocates distributed space and copy from (sequential) data */
    dist_list(int size, const A* seq_data);

    /** @brief A copy constructor
     *
     * Actually, this constructor does "shallow copy", that is, the two lists (this and that)
     * share the actual buffer on memary.  Use clone() in addition if you want "deep copy".
     *
     * @param that the reference to another distributed list
     */
    dist_list(const dist_list<A> &that);

    ~dist_list();

    /** @brief Assignement
     *
     * Actually, this assignment does "shallow copy", that is, the two lists (this and that)
     * share the actual buffer on memary.  Use clone() in addition if you want "deep copy".
     *
     * @param that the reference to another distributed list
     */
    dist_list<A>& operator=(const dist_list<A>& that);

  public:
    /** @brief Return the global size of the list */
    int get_global_size() const;

    /** @brief Return the size of elements allocated to the process */
    int get_local_size() const;
    
    /** @brief Get the value of the specified global index.
     * @param index the global index
     */
    A get(int index) const;

    /** @brief For skeleton implementation.
     * @param index the local index */
    A local_get(int index) const;

    /** @brief Set a value to the specifed global index.
     * @param index the global index.
     * @param val   value to be set.
     */
    void set(int index, A val);

    /** @brief Create a (deep) copy of the list. */
    dist_list<A> clone() const;

    /** @brief Gather the distributed values to the (sequential) array.
     *
     * This function gathers the distributed values to the array specified by the input.
     * The user should prepare enough space (i.e., get_global_size()).
     *
     * For reasons of performance, user can specify the buffer as NULL on some processes,
     * then this function does not gather the values to the process.
     * For example, this is effective when the gathered values are only used in outputting them
     * from the root process.
     *
     * @param buffer The pointer to the array to which the values are gathered, or NULL.
     */
    void gather(A* buffer) const;

#ifndef __SKETO_NO_FUSION__
    // The following are used when fusion transformation is activated.
  public:
    template <typename B>
    dist_list(const B& that);
    
    template <typename B>
    dist_list<A>& operator=(const B& that);

    template <typename T>
    void check_overwrite(_impl::dist_list_buffer<T>* target,
 			 int *refcount, bool *, bool *, bool *scan_shift) const
    {
      if ((void*)data == (void*)target) (*refcount)++;
      // *no_deps = true; *child_no_deps = ---;
      *scan_shift = false;
    }
      
#endif // __SKETO_NO_FUSION__
  };

  namespace _impl {

    /**
     * @brief The main data structure for maintaining distributed list.
     */
    template <typename A>
    class dist_list_buffer {
      friend class dist_list<A>;
      friend class _impl::ls_adpt;

    private:
      int ref_count;

      int global_size;
      A* buf;
      
      int local_begin;
      int local_size;

    public:
      dist_list_buffer(int gsize);
      ~dist_list_buffer();

      int get_global_size() const { return global_size; }
      int get_local_size() const { return local_size; }

      /** @brief Increment the reference count */
      void increment_ref_count();

      /** @brief Decrement the reference count and free the allocated memory when the count becomes zero */
      void decrement_ref_count();

    public:      // utility functions
      /**
       * @brief Compute the first local index and the local size of a distributed list.
       * @param global_size  global size of the list
       * @param procs        the number of processes
       * @param rank         the rank of process
       * @param local_begin  first local index will be stored
       * @param local_size   local size will be stored
       */
      static void
      compute_local_array(int global_size, int procs, int rank,
			  int* local_begin, int* local_size);

      /**
       * @brief Compute the rank of process that has the value of specified index.
       * @param global_size  global size of the list
       * @param procs        the number of processes
       * @param index        the index to be checked.
       */
      static int
      compute_rank_from_index(int global_size, int procs, int index);
    };

  }

}

#endif // __DIST_LIST_H__
