/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2009, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file dist_list.tpp
 * @brief Implementation of templete functions/classes.
 */
#ifndef __DIST_LIST_TPP__
#define __DIST_LIST_TPP__

#include <cassert>
#include <cstring>

namespace sketo {
  //------------------------------------------------------------------------
  // class dist_list

  template <typename A>
  dist_list<A>::dist_list()
    : data(NULL)
  {
  }

  template <typename A>
  dist_list<A>::dist_list(int size)
  {
    data = new _impl::dist_list_buffer<A>(size);
  }

  template <typename A>
  dist_list<A>::dist_list(int size, const A* seq_data)
  {
    data = new _impl::dist_list_buffer<A>(size);
    A* it = data->buf;
    int li, ls; _impl::dist_list_buffer<A>::compute_local_array(size, sketo::procs, sketo::rank, &li, &ls);
    const A* jt = seq_data + li; const A* const je = seq_data + li + ls;
    for (; jt != je; ++it, ++jt) {
      *it = *jt;
    }
  }

  template <typename A>
  dist_list<A>::dist_list(const dist_list<A> &that)
    : data(that.data)
  {
    if (data) data->increment_ref_count();
  }

  template <typename A>
  dist_list<A>::~dist_list()
  {
    if (data) data->decrement_ref_count();
  }

  template <typename A>
  dist_list<A> &dist_list<A>::operator=(const dist_list<A>& that)
  {
    // Do nothing if the input is the same as the output.
    if (this->data == that.data) return *this;

    if (data) data->decrement_ref_count();
    data = that.data;
    if (data) data->increment_ref_count();
    return *this;
  }

  template <typename A>
  inline int
  dist_list<A>::get_global_size() const
  {
    assert(data);
    return data->get_global_size();
  }

  template <typename A>
  inline int
  dist_list<A>::get_local_size() const
  {
    assert(data);
    return data->get_local_size();
  }

  template <typename A>
  A
  dist_list<A>::get(int index) const
  {
    A ret;
    int target_p = _impl::dist_list_buffer<A>::compute_rank_from_index(get_global_size(), sketo::procs, index);
    if (sketo::rank == target_p) {
      int lb, ls; _impl::dist_list_buffer<A>::compute_local_array(get_global_size(), sketo::procs, target_p, &lb, &ls);
      ret = data->buf[index - lb];
    }
    MPI_Bcast(&ret, sizeof(A), MPI_BYTE, target_p, MPI_COMM_WORLD);
    return ret;
  }

  template <typename A>
  inline A
  dist_list<A>::local_get(int index) const
  {
    return data->buf[index];
  }
  
  template <typename A>
  void
  dist_list<A>::set(int index, A val)
  {
    int target_p = _impl::dist_list_buffer<A>::compute_rank_from_index(data->global_size, sketo::procs, index);
    if (sketo::rank == target_p) {
      int lb, ls; _impl::dist_list_buffer<A>::compute_local_array(get_global_size(), sketo::procs, target_p, &lb, &ls);
      data->buf[index - lb] = val;
    }
  }

  template <typename A>
  dist_list<A>
  dist_list<A>::clone() const
  {
    dist_list<A> ret(get_global_size());

    A* const ret_data = ret.data->buf;
    const int n = get_local_size();
    for (int i = 0; i < n; ++i) {
      ret_data[i] = local_get(i);
    }

    return ret;
  }

  template <typename A>
  void
  dist_list<A>::gather(A* buffer) const
  {
    char nonnull = (buffer) ? 1 : 0;
    char* nonnulls = new char[sketo::procs];
    MPI_Allgather(&nonnull, sizeof(char), MPI_BYTE, nonnulls, sizeof(char), MPI_BYTE, MPI_COMM_WORLD);

    for (int p = 0; p < sketo::procs; ++p) {
      int lb, ls; _impl::dist_list_buffer<A>::compute_local_array(get_global_size(), sketo::procs, p, &lb, &ls);
      if (p == sketo::rank) {
	if (nonnull) std::memmove(buffer + lb, data->buf, sizeof(A)*ls);
	for (int pp = 0; pp < sketo::procs; ++pp) {
	  if (pp == sketo::rank) continue;
	  if (nonnulls[pp])
	    MPI_Send(static_cast<A*>(data->buf), sizeof(A)*ls, MPI_BYTE, pp, _impl::TAG_GATHER, MPI_COMM_WORLD);
	}
      } else {
	if (nonnull)
	  MPI_Recv(buffer + lb, sizeof(A)*ls, MPI_BYTE, p, _impl::TAG_GATHER, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      }
    }
  }

#ifndef __SKETO_NO_FUSION__
  template <typename A> template <typename B>
  dist_list<A>::dist_list(const B& that)
  {
    data = NULL;
    this->operator=(that);
  }
  
  template <typename A>
  template <typename B>
  dist_list<A> &dist_list<A>::operator=(const B& that)
  {
    // Condition for overwriting:
    //   - The sizes of lists are the same
    //   - Reference count is 1
    //   - No shift operations on the data
    int refcount = 0;
    bool no_deps = true;
    bool child_no_deps = true;
    bool scan_shift = false;
    that.check_overwrite(data, &refcount, &no_deps, &child_no_deps, &scan_shift);
    
    if ( data
	 && this->get_global_size() == that.get_global_size()
	 && data->ref_count - refcount == 1
	 && ((scan_shift && child_no_deps)
	      || (!scan_shift && no_deps))) {
      // do nothing.
    } else {
      // allocate new buffer
      if (data) data->decrement_ref_count();
      data = new _impl::dist_list_buffer<A>(that.get_global_size());
    }

    if (scan_shift && child_no_deps) {
      return *(that.ow_impl(this));
    } else {
      A* const as_buf = data->buf;
      const int n = that.get_local_size();
      for (int i = 0; i < n; ++i) {
	as_buf[i] = that.local_get(i);
      }
      return *this;
    }
  }
#endif
  

  //------------------------------------------------------------------------
  // class dist_list_buffer
  template <typename A>
  _impl::dist_list_buffer<A>::dist_list_buffer(int gsize)
    : ref_count(1), global_size(gsize)
  {
    compute_local_array(gsize, sketo::procs, sketo::rank, &local_begin, &local_size);
    buf = new A[local_size];
  }
  
  template <typename A>
  _impl::dist_list_buffer<A>::~dist_list_buffer()
  {
    assert(ref_count == 0);
  }

  template <typename A>
  void
  _impl::dist_list_buffer<A>::increment_ref_count()
  {
    assert(ref_count > 0);
    ++ref_count;
  }

  template <typename A>
  void
  _impl::dist_list_buffer<A>::decrement_ref_count()
  {
    assert(ref_count > 0);
    --ref_count;
    if (ref_count == 0) {
      delete [] buf; buf = NULL;
    }
  }

  template <typename A>
  void
  _impl::dist_list_buffer<A>::compute_local_array(int global_size,
						  int procs,
						  int rank,
						  int* local_begin,
						  int* local_size)
  {
    const int div = global_size / procs;
    const int mod = global_size % procs;
    if (rank >  mod)    { *local_begin = div * rank + mod;  *local_size = div;     return; }
    if (rank == mod)    { *local_begin = (div + 1) * rank;  *local_size = div;     return; }
    /* (rank <  mod) */ { *local_begin = (div + 1) * rank;  *local_size = div + 1; return; }
  }

  template <typename A>
  int
  _impl::dist_list_buffer<A>::compute_rank_from_index(int global_size,
						      int procs,
						      int index)
  {
    const int div = global_size / procs;
    const int mod = global_size % procs;
    int p = 0;
    int begin = 0;
    int next = div + (p >= mod ? 0 : 1);
    for (; p < procs; ++p) {
      if (begin <= index && index < next) break;
      begin = next; next += div + (p + 1 >= mod ? 0 : 1);
    }
    return p;
  }

}

#endif // __DIST_LIST_TPP__
