/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2009, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file nqueen.cpp
 * @brief This program compute N-Queens problem.
 * @author: Hideo NISHIMURA (nish@ipl.cs.uec.ac.jp)
 * @author: Kiminori Matsuzaki (kmatsu@ipl.t.u-tokyo.ac.jp; converted to new sketo implementation)
 */

//-----------------------------//
//         How to use          //
//-----------------------------//
//
// This program is solving the N-Queens problem using list skeletons.
//
// The N-Queens is a problem to place N queens on an NxN chess board such
// that no queen can attack another.
// 
//  Argument of console command :  %s N
//     "N" is board size
//

//-----------------------------//
//           Compile           //
//-----------------------------//
//
//  put C++ parallel library and "nqueen.cpp" in the same directory
//
//     $ mpiCC nqueen.cpp
//
//  to execute
//     $ mpirun -np "N" a.out "N"

#include <sketo/list_skeletons.h>
#include <cstdlib>

//---------------------------------//
//  functions using with Skeleton  //
//---------------------------------//

// --- recursive function solving N-Queen problem by backtrack algorithm
//   **argument
//       COUNT : the number of boards which fills a requirement ( output )
//         y   : level of the row height ( input )
//       left  : block branch to left direction ( input )
//       down  : block branch to down direction ( input )
//       right : block branch to right direction ( input )
//       size  : size of the board ( input )
inline void backtrack(int &COUNT, int y, int left, int down, int right, int size) {
  // all 1 bit which length is equal to board size
  const int mask = (1 << size) - 1;
  
  // when reaches to the bottom row of the board
  if ( y == size ) {
    COUNT++;
  } else {
    int bitmap = mask & ~(left | down | right);
    // backtracking
    while ( bitmap ) {
      const int bit = - bitmap & bitmap;
      bitmap ^= bit;
      backtrack(COUNT, y + 1, (left | bit) << 1, down | bit, (right | bit) >> 1, size);
    }
  }
}

// --- Solving N-Queens Problem
// --- Beforehand, set positions of 1st and 2nd row
// --- then, solve the board to be.
struct Solver : public sketo::functions::base <int (int)> {
  int size;
  Solver(int s) : size(s) {}
  
  int operator()(int index) const {
    int first  = 1;   // position of the 1st row
    int second = 1;  // position of the 2nd row

    // In case queen of the 1st low locates on the edge
    if (index < size - 2) {
      second <<= 2 + index;
    } else {
      // other case
      // the number of shift times in each row
      int f_shift = (index - 1) / (size - 3);
      int s_shift = (index - 1) % (size - 3);

      // In case a position of the 1st row ( and their shifted )
      // overlap a position of the 2nd row
      if (s_shift >= f_shift - 1) {
	s_shift += 3;
      }
      
      // shift "1" to proper position
      first  <<= f_shift;
      second <<= s_shift;
    }

    // block branches of each direction
    int d = first | second;
    int l = (first << 2) | (second << 1);
    int r = (first >> 2) | (second >> 1);

    // store the number of the boards fill the requirement
    int COUNT = 0;
      
    // Queen kernel
    backtrack(COUNT, 2, l, d, r, size);

    // the size of dist_list 
    const int n_size = ((size / 2) + (size % 2)) * (size - 3) + 1;
    
    // If queen position of 1st row doesn't locate the center of
    // the board, check half of the board and double it. 
    return ((size % 2) && (n_size - (size - 3) <= index)) ? COUNT : (2 * COUNT);
  }
};

//------------------------//
//     main function      //
//------------------------//
int sketo::main(int argc, char **argv)
{
  if (argc != 2) {
    sketo::cout << "usage: " << argv[0] << " N" << std::endl;
    return 1;
  }

  const int size = std::atoi(argv[1]); // board size ( i.e. N )

  // board size must be no less than 4
  if (size < 4) {
    sketo::cout << "board size range: 4-" << std::endl;
    return 1;
  }

//   skeleton::time::initialize( ); // for time keeping

  // size of dist_list 
  // ( expand up to 2nd row
  //   count half of the board, and double it )
  const int n_size = ((size / 2) + (size % 2)) * (size - 3) + 1;

  // solving N-Queens when constructing
  Solver solver(size);
  int total = list_skeletons::reduce(std::plus<int>(),
				     list_skeletons::generate(n_size, solver));

//   skeleton::time::set_time( "calculate" ); // for time keeping

  // output results
  sketo::cout << "SIZE: " << size << "\tTOTAL: " << total << std::endl;
//   skeleton::time::print( ); // for time keeping

  return 0;
}
